import { metrics, trace } from '@opentelemetry/api';
import { diag, DiagConsoleLogger, DiagLogLevel } from '@opentelemetry/api';
import { LogRecord, logs, SeverityNumber } from '@opentelemetry/api-logs';
import { getNodeAutoInstrumentations } from '@opentelemetry/auto-instrumentations-node';
import { OTLPLogExporter } from '@opentelemetry/exporter-logs-otlp-http';
import { OTLPMetricExporter } from '@opentelemetry/exporter-metrics-otlp-http';
import { OTLPTraceExporter } from '@opentelemetry/exporter-trace-otlp-http';
import { Resource } from '@opentelemetry/resources';
import { BatchLogRecordProcessor } from '@opentelemetry/sdk-logs';
import { ExponentialHistogramAggregation, MeterProvider, PeriodicExportingMetricReader, View } from '@opentelemetry/sdk-metrics';
import { NodeSDK } from '@opentelemetry/sdk-node';
import { SemanticResourceAttributes } from '@opentelemetry/semantic-conventions';
import { config } from 'dotenv';
import express, { Response } from 'express';
config();

const expHistogramView = new View({
	aggregation: new ExponentialHistogramAggregation(),
	// Note, the instrumentName is the same as the name that has been passed for
	// the Meter#createHistogram function for exponentialHistogram.
	instrumentName: 'test_exponential_histogram',
});

//
// For troubleshooting, set the log level to DiagLogLevel.DEBUG
const loggerDebugger = (LogLevel: DiagLogLevel) => diag.setLogger(new DiagConsoleLogger(), LogLevel);

const env = {
	service: {
		name: process.env.SERVER_HOSTNAME || 'undefined service',
		version: process.env.npm_package_version || 'undefined service version',
	},
	collector: {
		url: process.env.OPENTELEMETRY_COLLECTOR_URL || 'http://otelcol:4318',
	},
};

const sdk = new NodeSDK({
	resource: new Resource({
		[SemanticResourceAttributes.SERVICE_NAME]: env.service.name,
		[SemanticResourceAttributes.SERVICE_VERSION]: env.service.version,
	}),
	views: [expHistogramView],
	traceExporter: new OTLPTraceExporter({
		url: `${env.collector.url}/v1/traces`,
	}),
	metricReader: new PeriodicExportingMetricReader({
		exporter: new OTLPMetricExporter({
			url: `${env.collector.url}/v1/metrics`,
		}),
		exportIntervalMillis: 1000,
	}),
	logRecordProcessor: new BatchLogRecordProcessor(
		new OTLPLogExporter({
			url: `${env.collector.url}/v1/logs`,
		})
	),
	instrumentations: [getNodeAutoInstrumentations()],
});

const meter = metrics.getMeter(env.service.name);

const requestCounter = meter.createCounter('requests', {
	description: 'Example of a Counter',
});

const upDownCounter = meter.createUpDownCounter('test_up_down_counter', {
	description: 'Example of a UpDownCounter',
});

const histogram = meter.createHistogram('test_histogram', {
	description: 'Example of a Histogram',
});

const exponentialHistogram = meter.createHistogram('test_exponential_histogram', {
	description: 'Example of an ExponentialHistogram',
});

const attributes = { pid: process.pid, environment: 'staging' };

setInterval(() => {
	requestCounter.add(1, attributes);
	upDownCounter.add(Math.random() > 0.5 ? 1 : -1, attributes);
	histogram.record(Math.random(), attributes);
	exponentialHistogram.record(Math.random(), attributes);
}, 1000);

const app = express();
/**
 * Open Telemtry Logging middleware
 */

const loggingMiddleware = app.use((req, res, next) => {
	try {
		const activeSpan = trace.getActiveSpan();
		const logger = logs.getLogger(env.service.name);

		if (activeSpan !== undefined && logger !== undefined) {
			const temp = res.json;

			interface emitParams {
				severityNumber: SeverityNumber;
				severityText: string;
				context: string;
				message: unknown;
			}

			const log = ({ severityNumber, severityText, context, message }: emitParams) => {
				const obj: LogRecord = {
					severityNumber,
					severityText,
					body: JSON.stringify({ context, message }),
				};
				console.log('--- LOG ---', obj);
				logger.emit(obj);
			};

			res.json = function (...args): Response<unknown, Record<string, unknown>> {
				log({
					severityNumber: SeverityNumber.INFO,
					severityText: 'INFO',
					context: 'express request log',
					message: {
						request: {
							headers: req.headers,
							body: req.body,
							query: req.query,
							params: req.params,
						},
						response: {
							body: args['0'],
						},
					},
				});

				return temp.apply(this, args);
			};
		}
	} catch (e) {
		console.log('--- ERR :( ---', e);
	}
	next();
});

const logger = logs.getLogger(env.service.name);

export { DiagLogLevel, logger, loggerDebugger, loggingMiddleware, sdk, trace };
